%{
/* This file is part of Mailfromd.
   Copyright (C) 2005, 2006, 2007, 2008, 2009 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define MF_SOURCE_NAME  MF_SOURCE_LEX
	
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>	
#include <sys/stat.h>
#include "mailfromd.h"
#include "gram.h"
#include "prog.h"

static struct locus locus; /* Current input location */
static struct locus start_locus; /* Location when the last state switch
				    occurred */
static struct obstack string_stk; /* Obstack for constructing string values */
static char *multiline_delimiter; /* End of here-document delimiter */ 
static size_t multiline_delimiter_len; /* Length of multiline_delimiter_len */
static int multiline_unescape;         /* Unescape here-document contents */
static int (*char_to_strip)(char);     /* Strip matching characters of each
					  here-document line */

static int
is_tab(char c)
{
	return c == '\t';
}
 
static int
is_space(char c)
{
	return c == '\t' || c == ' ';
}

#define line_begin string_begin
#define line_add string_add
#define line_add_char string_add_char
static void line_finish(void);
static void string(const char *str, size_t len);
static int isemptystr(char *text);
static int builtin_const(const char *s, size_t len);

/* External interface to locus */ 
const struct locus *
get_locus()
{
	return &locus;
}

/* Auxiliary macro for returning keyword tokens */ 
#define keyword(kw) (yylval.locus = locus,(kw))


/* Input context flags */
#define INCTX_MODULE   0x1  /* current input is a MFL module */
#define INCTX_HADINPUT 0x2  /* some statements has already been processed */

/* Input context stack */
struct inctx {                     /* input context structure */
	struct inctx *parent;      /* parent context */
	struct locus locus;        /* locus where the context was pushed */
	struct input_file_ident id;/* file id structure to prevent recursion */
	FILE *file;                /* saved yyin */ 
	pid_t pp_pid;              /* preprocessor pid, if used */
	int inctx_flags;           /* input context flags */ 
	YY_BUFFER_STATE buf;       /* lex buffer state */
};

struct inctx *inctx_tos;           /* stack of input contexts */

/* Current input context: */
static pid_t pp_pid;               /* preprocessor pid */
static struct input_file_ident input_file_id;
static int inctx_flags;            /* input context flags */

/* If not 0, emit_token keeps a token that yylex must return on the
   next call. See YY_USER_ACTION below. */
static int emit_token;

/* Find on stack an input context that matches the given file id.
   Return pointer to the context. */
struct inctx *
inctx_locate(struct input_file_ident *id)
{
	struct inctx *ctx;

	for (ctx = inctx_tos; ctx; ctx = ctx->parent)
		if (ctx->id.device == id->device
		    && ctx->id.i_node == id->i_node)
			break;
	return ctx;
}
	
/* Push input context */	
static void
inctx_push()
{
	struct inctx *ctx = xmalloc(sizeof(*ctx));

	ctx->locus = locus;
	ctx->id = input_file_id;
	ctx->file = yyin;
	ctx->pp_pid = pp_pid;
	ctx->inctx_flags = inctx_flags;
	ctx->buf = YY_CURRENT_BUFFER;
	ctx->parent = inctx_tos;
	inctx_tos = ctx;
}

/* Pop input context from the top of the stack into the current
   input context. Return 1 if there are no more contexts left. */
int
inctx_pop()
{
	struct inctx *ctx = inctx_tos;
	
	if (!ctx)
		return 1;

	inctx_tos = ctx->parent;
	
	locus = ctx->locus;
	yyin = ctx->file; 
	input_file_id = ctx->id;
	inctx_flags = ctx->inctx_flags;
	yy_delete_buffer(YY_CURRENT_BUFFER);
	yy_switch_to_buffer(ctx->buf);
	pp_pid = ctx->pp_pid;
	free(ctx);
	return 0;
}

/* Setup the scanner for input from the file NAME. Return 0 on success,
   and an appropriate EX_ code on error. See lex_new_source below. */
static int
lex_new_source_0(const char *name)
{
	if (ext_pp) {
		char *ppcmd;
		
		yyin = pp_extrn_start(name, &pp_pid);
		if (!yyin) {
			parse_error(_("unable to start external preprocessor `%s': %s"),
				    ppcmd,
				    mu_strerror(errno));
			return EX_OSFILE;
		}
	} else {
		yyin = fopen(name, "r");
		if (!yyin) {
			parse_error(_("cannot open %s: %s"), name,
				    mu_strerror(errno));
			return EX_NOINPUT;
		}
	}
	yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
	assign_locus(&locus, name, "1");
	return EX_OK;
}

/* Save the current input context on stack and set up the scanner for
   input from the file NAME. Return 0 on success,
   and an appropriate EX_ code on error. */
int
lex_new_source(const char *name, int flag)
{
	int rc;
	struct stat st;
	struct input_file_ident id;
	struct inctx *pctx;
	
	if (stat(name, &st)) {
		parse_error(_("cannot open `%s': %s"),
			    name, mu_strerror(errno));
		return EX_NOINPUT;
	}
	
	id.device = st.st_dev;
	id.i_node = st.st_ino;
	
	pctx = inctx_locate(&id);
	if (pctx) {
		parse_error(_("Recursive inclusion"));
		if (pctx->parent)
			parse_error_locus(&pctx->parent->locus,
					  _("`%s' already included here"),
					  name);
		else
			parse_error(_("`%s' already included at top level"),
				    name);
		return 1;
	}
	
	if (flag == LEX_ONCE && source_lookup(&id))
		return 0;

	if (yyin)
		inctx_push();
	rc = lex_new_source_0(name);
	if (rc)
		inctx_pop();
	else {
		input_file_id = id;
		if (flag == LEX_MODULE) {
			inctx_flags = INCTX_MODULE;
			emit_token = T_MODBEG;
		} else
			inctx_flags = 0;
	}
	return rc;
}


/* Return constant or variable token corresponding to the current
   value of yylval.literal->text. */
static int
variable_or_const()
{
	struct variable *vptr;
	const struct constant *cptr;

	cptr = constant_lookup(yylval.literal->text);
	if (cptr) {
		const struct value *value_ptr = &cptr->value;
		switch (value_ptr->type) {
		case dtype_number:
			yylval.number = value_ptr->v.number;
			return T_NUMBER;

		case dtype_string:
			yylval.literal = value_ptr->v.literal;
			return T_STRING;
			
		default:
			abort();
		}
	}

	vptr = variable_lookup(yylval.literal->text);
	if (!vptr) {
		parse_error(_("Variable %s is not defined"),
			    yylval.literal->text);
		return T_BOGUS;
	}		
	yylval.var = vptr;
	return T_VARIABLE;
}
	
/* Advance locus to the next line */
void
advance_line()
{
	++locus.line;
}

/* Same as BEGIN, but also saves the current locus. It is then used
   to report unclosed constructs at the end of file. */
#define BEGIN_X(s) do { BEGIN(s); start_locus = locus; } while(0)

/* If emit_token is not 0, push back current input and return the value
   of emit_token. Clear emit_token before returning. */
#define YY_USER_ACTION							\
	if (emit_token) {						\
		int tok = emit_token;					\
		emit_token = 0;						\
		if (yy_flex_debug)					\
			fprintf(stderr, "--emitting %d (module %s)\n",	\
				tok,top_module->name);			\
		yyless(0);						\
		return tok;						\
	}								\

/* Read next input chunk. */
#define YY_INPUT(buf,result,max_size)					\
	if (yyin == NULL)						\
		result = YY_NULL;					\
	else if (((result = fread(buf, 1, max_size, yyin)) == 0)	\
		 && ferror(yyin))					\
		YY_FATAL_ERROR("input in flex scanner failed");

/* Redeclare main entry point. Actual yylex is defined in the code section
   below. */
#define YY_DECL static int lexscan(void)
%}

/* Exclusive states:

   COMMENT        Within a C-style comment;
   STR            Processing a complex string;
   ML             Within a multi-line aggregator.  The line being built
                  requires stripping leading whitespace (if requested).
   CML            Continuation within a multi-line aggregator.  The line
                  being built does not require stripping leading whitespace.
   QML            Quoted multi-line aggregator.  No variable substitution and
                  unquoting is needed.

   Inclusive states:
   
   ONBLOCK	  Lexical tie-in after an `on' keyword.  In ONBLOCK state
                  the strigns `as', `host', `from', and `poll' are
		  recognized as keywords.  The string `for' also acquires
		  special meaning.
*/

%x COMMENT STR ML CML QML
%s ONBLOCK

N [0-9][0-9]*
P [1-9][0-9]*
X [0-9a-fA-F]
O [0-7]
WS [ \t][ \t]*
IDENT [a-zA-Z_][a-zA-Z_0-9]*
LOCUS __file__|__line__|__function__|__module__
VCONST __package__|__version__|__major__|__minor__|__patch__
STATEDIR __(def)?statedir__
PREPROC  __(def)?preproc__	
ICONST {LOCUS}|{VCONST}|{STATEDIR}|{PREPROC}
%%
         /* C-style comments */
"/*"         BEGIN_X(COMMENT);
<COMMENT>[^*\n]*        /* eat anything that's not a '*' */
<COMMENT>"*"+[^*/\n]*   /* eat up '*'s not followed by '/'s */
<COMMENT>\n             advance_line();
<COMMENT>"*"+"/"        BEGIN(INITIAL);
         /* Configuration directives */
^[ \t]*#[ \t]*{P}[ \t]+\".*\".*\n parse_line_cpp(yytext, &locus);
        /* Normally, everything within a comment should be ignored, so
	   the exclusive condition for the rule below is an error.
	   Unfortunately, GNU m4 (versions up to 1.4.9) outputs line
	   synchronisation directives in comments, which makes any decent
	   compiler lost trace of which input line it is on.  To avoid
	   this, mfd handles #line directives even within a C-comment
	   block.

	   The bug report with a patch was sent to bug-m4@gnu.org on
	   2007-05-24:

	     http://lists.gnu.org/archive/html/bug-m4/2007-05/msg00018.html
	     
	   Hopefully it will be fixed in the next release.
	*/
<INITIAL,COMMENT>^[ \t]*#[ \t]*line[ \t].*\n {
	parse_line(yytext, &locus);
}
^[ \t]*#[ \t]*pragma[ \t].*/\n parse_pragma(yytext);
^[ \t]*#[ \t]*error[ \t].*\n  {
         yytext[yyleng-1] = 0; /* Kill trailing newline */
         parse_error("%s", yytext);
         advance_line();
 }
^[ \t]*#[ \t]*include_once[ \t].*\n {
	parse_include(yytext, 1);
}
^[ \t]*#[ \t]*include[ \t].*\n {
	parse_include(yytext, 0);
}
^[ \t]*#[ \t]*require[ \t].*\n {
	deprecation_warning(_("#require is deprecated, use `require' or `from ... import' instead"));
	/* parse_require cannot be called directly from lexer to avoid
	   switching module context in the middle of grammar rule. */
	yylval.string = yytext;
	return T_SHARPREQR;
}
         /* End-of-line comments */
#.*\n   { advance_line(); }
#.*     /* end-of-file comment */;
         /* Reserved words */
accept    return keyword(T_ACCEPT);
reject    return keyword(T_REJECT);
tempfail  return keyword(T_TEMPFAIL);
continue  return keyword(T_CONTINUE);
discard   return keyword(T_DISCARD);
add       return keyword(T_ADD);
replace   return keyword(T_REPLACE);
delete    return keyword(T_DELETE);
if        return keyword(T_IF);
fi        return keyword(T_FI);
else      return keyword(T_ELSE);
elif      return keyword(T_ELIF);
on        return keyword(T_ON);
do        return keyword(T_DO);
done      return keyword(T_DONE);
matches   return keyword(T_MATCHES);
fnmatches return keyword(T_FNMATCHES);
mx{WS}matches return keyword(T_MXMATCHES);
mx{WS}fnmatches return keyword(T_MXFNMATCHES);
when      return keyword(T_WHEN);
or        return keyword(T_OR);
and       return keyword(T_AND);
not       return keyword(T_NOT);
next      return keyword(T_NEXT);
prog      return keyword(T_PROG);
set       return keyword(T_SET);
catch     return keyword(T_CATCH);
echo      return keyword(T_ECHO);
return    return keyword(T_RETURN);
returns   return keyword(T_RETURNS);
func      return keyword(T_FUNC);
switch    return keyword(T_SWITCH);
case      return keyword(T_CASE);
default   return keyword(T_DEFAULT);
string    { yylval.type = dtype_string; return T_TYPE; }
number    { yylval.type = dtype_number; return T_TYPE; }
const     return keyword(T_CONST);
throw     return keyword(T_THROW);
loop      return keyword(T_LOOP);
while     return keyword(T_WHILE);
for       return keyword(T_FOR);
break     return keyword(T_BREAK);
pass      return keyword(T_PASS);
begin     return keyword(T_BEGIN);
end       return keyword(T_END);
alias     return keyword(T_ALIAS);
vaptr     return keyword(T_VAPTR);
precious  return keyword(T_PRECIOUS);
require   return keyword(T_REQUIRE);
import    return keyword(T_IMPORT);
from      return keyword(T_FROM);
static    return keyword(T_STATIC);
public    return keyword(T_PUBLIC);
module    {  if (inctx_tos && inctx_flags == 0) {
		parse_warning_locus(&inctx_tos->locus,
				    _("including a module file is unreliable and may cause subtle errors"));
		/* TRANSLATORS: Do not translate `require %s' */
		parse_warning_locus(&inctx_tos->locus,
				    _("use `require %s' instead"),
				    locus.file);
	     } else if (inctx_flags & INCTX_HADINPUT)
		     parse_error(_("misplaced `module'"));
	    return keyword(T_MODULE);
          }
bye       return keyword(T_BYE);

{ICONST}   { return builtin_const(yytext, yyleng); } 
<ONBLOCK>poll      return keyword(T_POLL);
<ONBLOCK>host      return keyword(T_HOST);
<ONBLOCK>as        return keyword(T_AS);
         /* Variables */
<INITIAL,ONBLOCK,ML,CML,STR>\%({ICONST}) {
	return builtin_const(yytext + 1, yyleng - 1);
}
<INITIAL,ONBLOCK,ML,CML,STR>\%\{({ICONST})\} {
	return builtin_const(yytext + 2, yyleng - 3);
}
<INITIAL,ONBLOCK,ML,CML,STR>\%{IDENT}     {
	string(yytext + 1, yyleng - 1);
	return variable_or_const();
}
<INITIAL,ONBLOCK,ML,CML,STR>\%\{{IDENT}\} {
	string(yytext + 2, yyleng - 3);
	return variable_or_const();
}

         /* Positional arguments */
<INITIAL,ONBLOCK,ML,CML,STR>\$# {
	return T_ARGCOUNT;
}
<INITIAL,ONBLOCK,ML,CML,STR>\${P}        {
	yylval.number = strtol(yytext + 1, NULL, 0);
	return T_ARG; }
<INITIAL,ONBLOCK>\$/"("        { return keyword(T_ARGX); }
         /* Sendmail variables */
<INITIAL,ONBLOCK,ML,CML,STR>\${IDENT} {
	        if (yyleng == 2)
	           string(yytext + 1, 1);
                else {
                   line_begin();
                   line_add("{", 1);
                   line_add(yytext + 1, yyleng - 1);
                   line_add("}", 1);
                   line_finish();
                }
                return T_SYMBOL;
              }
<INITIAL,ONBLOCK,ML,CML,STR>\$\{{IDENT}\} {
	string(yytext+1, yyleng - 1); return T_SYMBOL; }
         /* Back-references */
<INITIAL,ONBLOCK,ML,CML,STR>\\{P} {
	 yylval.number = strtoul(yytext+1, NULL, 0);
	 return T_BACKREF; }
         /* Numeric strings */
0[xX]{X}{X}*  { yylval.number = strtoul(yytext, NULL, 16); return T_NUMBER; };
0{O}{O}*      { yylval.number = strtoul(yytext, NULL, 8); return T_NUMBER; };
0|{P}         { yylval.number = strtoul(yytext, NULL, 10); return T_NUMBER; };
         /* Strings */
{IDENT}  {
	int paren_follows = 0;
	enum { is_ident, is_builtin, is_func } state;

	if (yylval.builtin = builtin_lookup(yytext))
		state = is_builtin;
	else if (yylval.function = function_lookup(yytext))
		state = is_func;
	else
		state = is_ident;

	if (state != is_ident) {
		int c;
		
		while ((c = input()) && c_isascii(c) && c_isspace(c))
			if (c == '\n')
				advance_line();
	
		paren_follows = (c == '(');
		unput(c);
	}
	
	if (state == is_builtin) 
		return yylval.builtin->rettype == dtype_unspecified ?
			T_BUILTIN_PROC :
			(paren_follows ? T_BUILTIN : T_BUILTIN_P);
	else if (state == is_func)
		return yylval.function->rettype == dtype_unspecified ?
			T_FUNCTION_PROC :
			(paren_follows ? T_FUNCTION : T_FUNCTION_P);
	else {
		string(yytext, yyleng);
		return T_IDENTIFIER;
	}
  }
'[^\n']*'             { string(yytext+1, yyleng-2); return T_STRING; }
\"[^\\\"$%\n]*\"      { string(yytext+1, yyleng-2); return T_STRING; }
\"[^\\\"$%\n]*\\\n    { advance_line();
                        BEGIN_X(STR);
                        line_begin();
                        line_add(yytext + 1, yyleng - 3); }     
\"[^\\\"$%\n]*/[\\$%] { BEGIN_X(STR);
                        string(yytext+1, yyleng-1);
                        line_begin();
                        return T_STRING; }
\"\\x{X}{X}/[\\$%] {
         BEGIN_X(STR);
         line_add_char(strtoul(yytext + 3, NULL, 16));
         line_finish();
         return T_STRING;
}
\"\\x{X}{X} {
         BEGIN_X(STR);
         line_add_char(strtoul(yytext + 3, NULL, 16));
}

\"\\0{O}{1,3}/[\\$%] {
         BEGIN_X(STR);  
         line_add_char(strtoul(yytext + 3, NULL, 8));
         line_finish();
         return T_STRING;
}
\"\\0{O}{1,3} {
         BEGIN_X(STR);  
         line_add_char(strtoul(yytext + 3, NULL, 8));
}

\"\\[^1-9]/[\\$%]  {
         BEGIN_X(STR);
         line_add_char(mu_argcv_unquote_char(yytext[2]));
         line_finish();
         return T_STRING;
}
\"\\[^1-9]  {
         BEGIN_X(STR);
         line_add_char(mu_argcv_unquote_char(yytext[2]));
}
<STR>[^\\\"$%\n]*\\\n { advance_line(); line_add(yytext, yyleng - 2); }
<STR>[^\\\"$%\n]*\"   { BEGIN(INITIAL);
                        if (yyleng > 1) 
                          line_add(yytext, yyleng - 1); 
                        line_finish();
		        return T_STRING; }
<STR>[^\\\"$%\n]+/[\\$%] {
                       line_add(yytext, yyleng);
                       line_finish();
                       line_begin();
                       return T_STRING;
                 }
<STR,ML,CML>\\x{X}{X}/[\\$%] {
         line_add_char(strtoul(yytext + 2, NULL, 16));
         line_finish();
         line_begin();
         return T_STRING;
}
<STR,ML,CML>\\x{X}{X} {
         line_add_char(strtoul(yytext + 2, NULL, 16));
}
<STR,ML,CML>\\0{O}{1,3}/[\\$%] {
         line_add_char(strtoul(yytext + 2, NULL, 8));
         line_finish();
         line_begin();
         return T_STRING;
}
<STR,ML,CML>\\0{O}{1,3} {
         line_add_char(strtoul(yytext + 2, NULL, 8));
}
<STR,ML,CML>\\[^1-9]/[\\$%] {
         line_add_char(mu_argcv_unquote_char(yytext[1]));
         line_finish();
         line_begin();
         return T_STRING;
}
<STR,ML,CML>\\[^1-9] {
         line_add_char(mu_argcv_unquote_char(yytext[1]));
}

         /* Multi-line strings */
"<<"(-" "?)?\\?{IDENT}[ \t]*.*\n |
"<<"(-" "?)?'{IDENT}'[ \t]*.*\n {
        char *p;

        advance_line();
        char_to_strip = NULL;
        multiline_unescape = 1;

        line_begin();
        p = yytext + 2;
        if (*p == '-') {
            ++p;
            if (*p == ' ') {
                ++p;
                char_to_strip = is_space;
            } else
                char_to_strip = is_tab;
        }
        
        if (*p == '\\') {
            p++;
            multiline_unescape = 0;
        }
        if (*p == '\'') {
            char *q;
 
            p++;
            multiline_unescape = 0;
            q = strchr(p, '\'');
            multiline_delimiter_len = q - p;
        } else
            multiline_delimiter_len = strcspn(p, " \t");

        multiline_delimiter = xmalloc(multiline_delimiter_len + 1);
        memcpy(multiline_delimiter, p, multiline_delimiter_len);
        multiline_delimiter[multiline_delimiter_len] = 0;
        if (multiline_unescape)
            BEGIN_X(ML);
        else
            BEGIN_X(QML);
}
        /* Quoted multilines */
<QML>[^\\\n]*\n {
        char *p;

        advance_line();
        p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;

        if (strlen(p) >= multiline_delimiter_len
            && memcmp(p, multiline_delimiter, multiline_delimiter_len) == 0
            && isemptystr(p + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return T_STRING;
        }
        line_add(p, strlen(p));
}

        /* Unquoted multilines */
<ML>[^\\$%\n]+/[\\$%] {
        char *p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;
        line_add(p, strlen(p));
        line_finish();
        BEGIN(CML);
        return T_STRING;
}
<CML>[^\\$%\n]+/[\\$%] {
        line_add(yytext, yyleng);
        line_finish();
        line_begin();
        return T_STRING;
}
<CML,STR>"%%"|"$$" {
        line_add(yytext, 1);
        line_finish();
        return T_STRING;
}
<ML,CML,STR>[$%] {
        line_add(yytext, yyleng);
}
<ML>[^\\$%\n]*\n/% {
        char *p;

        advance_line();
        p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;

        if (strlen(p) >= multiline_delimiter_len
            && memcmp(p, multiline_delimiter, multiline_delimiter_len) == 0
            && isemptystr(p + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return T_STRING;
        }
        line_add(p, strlen(p));
        line_finish();
        return T_STRING;
}
<ML>[^\\$%\n]*\n {
        char *p;

        advance_line();
        p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;

        if (strlen(p) >= multiline_delimiter_len
            && memcmp(p, multiline_delimiter, multiline_delimiter_len) == 0
            && isemptystr(p + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return T_STRING;
        }
        line_add(p, strlen(p));
}
<CML>[^\\$%\n]*\n {
        advance_line();
        if (yyleng >= multiline_delimiter_len
            && memcmp(yytext, multiline_delimiter,
                      multiline_delimiter_len) == 0
            && isemptystr(yytext + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return T_STRING;
        }
        line_add(yytext, yyleng);
        BEGIN_X(ML);
}
         /* Other tokens */
{WS}     ;
\n { advance_line(); }
"="|"==" return keyword(T_EQ);
"!=" return keyword(T_NE);
"<"  return keyword(T_LT);
"<=" return keyword(T_LE);
">"  return keyword(T_GT);
">=" return keyword(T_GE);
"&"  return keyword(T_LOGAND);
"|"  return keyword(T_LOGOR);
"^"  return keyword(T_LOGXOR);
"~"  return keyword(T_LOGNOT);
"..." return keyword(T_DOTS);
<ML,QML>.   /* If a here-document is not closed and its next line does not
               end with a \n, prevent it from being displayed by ECHO */;
. return yytext[0];
%%

int
yylex()
{
        int rc = lexscan();
	if (rc != T_MODBEG && rc != T_MODEND)
		inctx_flags |= INCTX_HADINPUT;
	return rc;
}

void
init_string_space()
{
	obstack_init(&string_stk);
}

void
free_string_space()
{
	obstack_free(&string_stk, NULL);
}

char *
mf_strdup(const char *str)
{
	string_add(str, strlen(str) + 1);
	return obstack_finish(&string_stk);
}

struct literal *
string_alloc(const char *str, size_t len)
{
	string_begin();
	string_add(str, len);
	return string_finish();
}

static void
string(const char *str, size_t len)
{
	yylval.literal = string_alloc(str, len);
}

void
string_begin()
{
	/* nothing */
}

struct literal *
string_finish()
{
	char *ptr;
	struct literal *lit;
	obstack_1grow(&string_stk, 0);
	ptr = obstack_finish(&string_stk);
	lit = literal_lookup(ptr);
	if (lit->text != ptr)
		obstack_free(&string_stk, ptr);
	return lit;
}

static void
line_finish()
{	
	yylval.literal = string_finish();
	if (yy_flex_debug)
		fprintf(stderr, "constructed line: %s\n",
			yylval.literal->text);
}		

void
string_add(const char *str, size_t len)
{
	obstack_grow(&string_stk, str, len);
}

void
string_add_char(unsigned char c)
{
	obstack_1grow(&string_stk, c);
}

static void
print_parse_message(const struct locus *locus,
		    const char *prefix, const char *fmt, va_list ap)
{
	int n;
	char buf[512]; /* FIXME */

	if (locus && locus->file)
		n = snprintf(buf, sizeof buf, "%s:%lu: ",
			     locus->file, (unsigned long) locus->line);
	else
		n = 0;
	if (prefix) 
		n += snprintf(buf + n, sizeof buf - n, "%s: ", prefix);
	vsnprintf(buf + n, sizeof buf - n, gettext(fmt), ap);
	mu_error(buf);
}

void
vparse_warning(const char *fmt, va_list ap)
{
	print_parse_message(&locus, _("warning"), fmt, ap);
}

void
parse_warning(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(&locus, _("warning"), fmt, ap);
	va_end(ap);
}

void
parse_warning_locus(const struct locus *loc, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(loc, _("warning"), fmt, ap);
	va_end(ap);
}

void
parse_error(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(&locus, NULL, fmt, ap);
	va_end(ap);
	error_count++;
}

void
parse_error_locus(const struct locus *loc, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(loc, NULL, fmt, ap);
	va_end(ap);
	error_count++;
}



static int
lex_close_source()
{
	if (!yyin)
		return 1;
	if (ext_pp)
 	        pp_extrn_shutdown(yyin, pp_pid);
	else
		fclose(yyin);
	yyin = NULL;
	
	switch (YYSTATE) {
	case INITIAL:
		/* ok */
		break;

	case COMMENT:
		parse_error_locus(&start_locus, _("end of file in comment"));
		break;

	case STR:
	case ML:
	case CML:
	case QML:
		parse_error_locus(&start_locus, _("end of file in string"));
	}
	return inctx_pop();
}

void
lex_drain_input()
{
	while (input() != EOF)
		;
}

void
lex_bye()
{
	lex_drain_input();
	lex_close_source();
}

int
yywrap()
{
	if (yy_flex_debug && top_module)
		fprintf(stderr, "--eof in module %s\n", top_module->name);
	if (inctx_flags & INCTX_MODULE)
		emit_token = T_MODEND;
	return lex_close_source();
}

static int
isemptystr(char *text)
{
	for (; *text && c_isspace (*text); text++)
		;
	return *text == 0;
}

void
tie_in_onblock(int enable)
{
	if (enable)
		BEGIN(ONBLOCK);
	else
		BEGIN(INITIAL);
}

data_type_t
builtin_const_value(const char *s, size_t len, const char **sval, long *nval)
{
	if (strncmp(s, "__file__", len) == 0) {
		struct literal *lit = literal_lookup(locus.file);
		*sval = lit->text;
		return dtype_string;
	} else if (strncmp(s, "__line__", len) == 0) {
		*nval = locus.line;
		return dtype_number;
	} else if (strncmp(s, "__function__", len) == 0) {
		*sval = function_name();
		return dtype_string;
	} else if (strncmp(s, "__package__", len) == 0) {
		*sval = PACKAGE_TARNAME;
		return dtype_string;
	} else if (strncmp(s, "__version__", len) == 0) {
		*sval = PACKAGE_VERSION;
		return dtype_string;
	} else if (strncmp(s, "__major__", len) == 0) {
		*nval = MAILFROMD_VERSION_MAJOR;
		return dtype_number;
	} else if (strncmp(s, "__minor__", len) == 0) {
		*nval = MAILFROMD_VERSION_MINOR;
		return dtype_number;
	} else if (strncmp(s, "__patch__", len) == 0) {
		*nval = MAILFROMD_VERSION_PATCH;
		return dtype_number;
	} else if (strncmp(s, "__statedir__", len) == 0) {
                *sval = mailfromd_state_dir;
		return dtype_string;
	} else if (strncmp(s, "__defstatedir__", len) == 0) {
		*sval = DEFAULT_STATE_DIR;
		return dtype_string;
	} else if (strncmp(s, "__preproc__", len) == 0) {
		*sval = ext_pp;
		return dtype_string;
	} else if (strncmp(s, "__defpreproc__", len) == 0) {
		*sval = (DEFAULT_PREPROCESSOR) ? DEFAULT_PREPROCESSOR : "";
		return dtype_string;
	} else if (strncmp(s, "__module__", len) == 0) {
		*sval = top_module->dclname ? top_module->dclname : top_module->name;
		return dtype_string;
	}
        return dtype_unspecified;  
}

int
builtin_const(const char *s, size_t len)
{
	const char *sval;
	long nval;
	
	switch (builtin_const_value(s, len, &sval, &nval)) {
	case dtype_number:
		yylval.number = nval;
		return T_NUMBER;

	case dtype_string:
		string(sval, strlen(sval));
		return T_STRING;

	default:
		abort();
	}
	return T_BOGUS;
}

const char *
symbit_to_qualifier(unsigned f)
{
        switch (f) {
	case SYM_PRECIOUS:
		return "precious";
	case SYM_STATIC:
		return "static";
	case SYM_PUBLIC:
		return "public";
	}
	return NULL;
}

/* End of lex.l */
